###
Docpad by Benjamin Lupton
Intuitive Web Development
###

# Requirements
fs = require 'fs'
path = require 'path'
sys = require 'util'
caterpillar = require 'caterpillar'
util = require 'bal-util'
child_process = require('child_process')
growl = require('growl')
express = false
watch = false
queryEngine = false
_ = require 'underscore'
PluginLoader = require "#{__dirname}/plugin-loader.coffee"
require "#{__dirname}/prototypes.coffee"
exec = (commands,options,callback) ->
	# Sync
	tasks = new util.Group callback
	
	# Tasks
	commands = [commands]  unless commands instanceof Array
	for command in commands
		console.log "Adding [#{command}]"
		tasks.push ->
			console.log "Running [#{command}] [#{tasks.queueIndex}]"
			child_process.exec command, options, tasks.completer()

	# Execute the tasks synchronously
	console.log require('util').inspect tasks.queue
	tasks.sync()


# -------------------------------------
# Docpad

class Docpad
	# Options
	rootPath: null
	outPath: 'out'
	srcPath: 'src'
	corePath: "#{__dirname}/.."
	libPath: "#{__dirname}"
	mainPath: "#{__dirname}/docpad.coffee"
	layoutsPath: null
	documentsPath: null
	publicPath: null
	skeleton: 'bootstrap'
	maxAge: false
	logLevel: 6

	# Docpad
	version: null
	packageData: null
	config: null
	generating: false
	loading: true
	generateTimeout: null
	actionTimeout: null
	server: null
	port: 9778
	logger: null

	# Models
	File: null
	Layout: null
	Document: null
	layouts: null
	documents: null
	
	# Plugins
	pluginsArray: []
	pluginsObject: {}

	# Skeletons
	skeletons:
		bootstrap:
			repo: 'https://github.com/balupton/bootstrap.docpad.git'


	# ---------------------------------
	# Main

	# Init
	constructor: ({rootPath,outPath,srcPath,skeleton,maxAge,port,server,logLevel,config}={}) ->
		# Options
		@config = {}
		@pluginsArray = []
		@pluginsObject = {}
		@port = port  if port
		@maxAge = maxAge  if maxAge
		@server = server  if server
		@rootPath = path.normalize(rootPath || process.cwd())
		@outPath = path.normalize(outPath || "#{@rootPath}/#{@outPath}")
		@srcPath = path.normalize(srcPath || "#{@rootPath}/#{@srcPath}")
		@layoutsPath = @srcPath+'/layouts'
		@documentsPath = @srcPath+'/documents'
		@publicPath = @srcPath+'/public'
		@skeleton = skeleton  if skeleton

		# Logger
		@logLevel = logLevel or @logLevel
		@logger = new caterpillar.Logger
			transports:
				level: @logLevel
				formatter:
					module: module
		
		# Configure
		@loadConfiguration(config)

		# Version Check
		@compareVersion()
		
		# Load Plugins
		@loadPlugins null, (err) =>
			@error(err)  if err
			@loading = false
	
	# Load Configuration
	loadConfiguration: (config={}) ->
		# Prepare
		@config =
			enableUnlistedPlugins: true
			enabledPlugins: {}
			plugins: {}

		# Docpad Configuration
		@packageData = JSON.parse(fs.readFileSync("#{@corePath}/package.json").toString()) or {}
		@version = @packageData.version
		_.extend @config, @packageData.docpad or {}

		# Website Configuration
		try
			websiteData = JSON.parse(fs.readFileSync("#{@rootPath}/package.json").toString()) or {}
			_.extend @config, websiteData.docpad or {}
		catch err
			@error('Failed to load the website configuration')
		
		# User Configuration
		_.extend @config, config
		
		# Prepare enabled plugins
		if typeof @config.enabledPlugins is 'string'
			enabledPlugins = {}
			for enabledPlugin in @config.enabledPlugins.split(/[ ,]+/)
				enabledPlugins[enabledPlugin] = true
			@config.enabledPlugins = enabledPlugins
		
		# Chain
		@
	
	# Layout Document
	createDocument: (meta={}) ->
		# Prepare
		config = docpad: @, layouts: @layouts, logger: @logger, meta: meta
		
		# Create and return
		document = new @Document config
	
	# Create Layout
	createLayout: (meta={}) ->
		# Prepare
		config = docpad: @, layouts: @layouts, logger: @logger, meta: meta

		# Create and return
		layout = new @Layout config

	# Clean Models
	cleanModels: (next) ->
		# Prepare
		File = @File = require("#{@libPath}/file.coffee")
		Layout = @Layout = class extends File
		Document = @Document = class extends File
		layouts = @layouts = new queryEngine.Collection
		documents = @documents = new queryEngine.Collection
		
		# Extend
		Layout::store = ->
			layouts[@id] = @
		Document::store = ->
			documents[@id] = @
		
		# Next
		next()  if next

		# Chain
		@

	# Compare versions
	compareVersion: ->
		util.packageCompare
			local: "#{@corePath}/package.json"
			remote: 'https://raw.github.com/balupton/docpad/master/package.json'
			newVersionCallback: (details) =>
				growl.notify 'There is a new version of docpad available'
				@logger.log 'notice', """
					There is a new version of docpad available, you should probably upgrade...
					current version:  #{details.local.version}
					new version:      #{details.remote.version}
					grab it here:     #{details.remote.homepage}
					"""
		@

	# Initialise the Skeleton
	initialiseSkeleton: (skeleton, destinationPath, next) ->
		# Prepare
		logger = @logger
		skeletonRepo = @skeletons[skeleton].repo
		logger.log 'info', "[#{skeleton}] Initialising the Skeleton to #{destinationPath}"
		snoring = false
		notice = setTimeout(
			->
				snoring = true
				logger.log 'notice', "[#{skeleton}] This could take a while, grab a snickers"
			2000
		)

		# Async
		tasks = new util.Group (err) ->
			logger.log 'info', "[#{skeleton}] Initialised the Skeleton"  unless err
			next(err)
		tasks.total = 2
		
		# Pull
		logger.log 'debug', "[#{skeleton}] Pulling in the Skeleton"
		child = exec(
			# Commands
			[
				"git init"
				"git remote add skeleton #{skeletonRepo}"
				"git pull skeleton master"
			]
			
			# Options
			{
				cwd: destinationPath
			}

			# Next
			(err,stdout,stderr) ->
				# Output
				if err
					console.log stdout.replace(/\s+$/,'')  if stdout
					console.log stderr.replace(/\s+$/,'')  if stderr
					return next(err)
				
				# Log
				logger.log 'debug', "[#{skeleton}] Pulled in the Skeleton"

				# Submodules
				path.exists "#{destinationPath}/.gitmodules", (exists) ->
					tasks.complete()  unless exists
					logger.log 'debug', "[#{skeleton}] Initialising Submodules for Skeleton"
					child = exec(
						# Commands
						[
							'git submodule init'
							'git submodule update'
							'git submodule foreach --recursive "git init"'
							'git submodule foreach --recursive "git checkout master"'
							'git submodule foreach --recursive "git submodule init"'
							'git submodule foreach --recursive "git submodule update"'
						]
						
						# Options
						{
							cwd: destinationPath
						}

						# Next
						(err,stdout,stderr) ->
							# Output
							if err
								console.log stdout.replace(/\s+$/,'')  if stdout
								console.log stderr.replace(/\s+$/,'')  if stderr
								return tasks.complete(err)  
							
							# Complete
							logger.log 'debug', "[#{skeleton}] Initalised Submodules for Skeleton"
							tasks.complete()
					)
				
				# NPM
				path.exists "#{destinationPath}/package.json", (exists) ->
					tasks.complete()  unless exists
					logger.log 'debug', "[#{skeleton}] Initialising NPM for Skeleton"
					child = exec(
						# Command
						'npm install'

						# Options
						{
							cwd: destinationPath
						}

						# Next
						(err,stdout,stderr) ->
							# Output
							if err
								console.log stdout.replace(/\s+$/,'')  if stdout
								console.log stderr.replace(/\s+$/,'')  if stderr
								return tasks.complete(err)  
							
							# Complete
							logger.log 'debug', "[#{skeleton}] Initialised NPM for Skeleton"
							tasks.complete()
					)
		)

		# Chain
		@
	
	# Handle
	action: (action,next=null) ->
		# Prepare
		logger = @logger
		next or= -> process.exit()

		# Clear
		if @actionTimeout
			clearTimeout(@actionTimeout)
			@actionTimeout = null

		# Check
		if @loading
			@actionTimeout = setTimeout(
				=>
					@action(action, next)
				1000
			)
			return
		
		# Handle
		switch action
			when 'skeleton', 'scaffold'
				@skeletonAction (err) ->
					return @error(err)  if err
					next()

			when 'generate'
				@generateAction (err) ->
					return @error(err)  if err
					next()

			when 'watch'
				@watchAction (err) ->
					return @error(err)  if err
					logger.log 'info', 'DocPad is now watching you...'

			when 'server', 'serve'
				@serverAction (err) ->
					return @error(err)  if err
					logger.log 'info', 'DocPad is now serving you...'

			else
				@skeletonAction (err) =>
					return @error(err)  if err
					@generateAction (err) =>
						return @error(err)  if err
						@serverAction (err) =>
							return @error(err)  if err
							@watchAction (err) =>
								return @error(err)  if err
								logger.log 'info', 'DocPad is now watching and serving you...'
		
		# Chain
		@
	

	# Handle an error
	error: (err) ->
		return  unless err
		@logger.log 'err', 'An error occured:', err.message, err.stack
		@




	# ---------------------------------
	# Plugins


	# Get a plugin by it's name
	getPlugin: (pluginName) ->
		@pluginsObject[pluginName]


	# Trigger a plugin event
	# next(err)
	triggerEvent: (eventName,data,next) ->
		# Async
		logger = @logger
		tasks = new util.Group (err) ->
			logger.log 'debug', "Plugins completed for #{eventName}"
			next(err)
		tasks.total = @pluginsArray.length

		# Cycle
		for plugin in @pluginsArray
			plugin.triggerEvent eventName, data, tasks.completer()
		
		# Chain
		@


	# Load Plugins
	loadPlugins: (pluginsPath, next) ->
		# Prepare
		logger = @logger
		docpad = @

		# Default Load
		unless pluginsPath
			# Async
			tasks = new util.Group (err) ->
				logger.log 'debug', 'All plugins loaded'
				next(err)
			
			# Load in the docpad and local plugin directories
			tasks.push => @loadPlugins "#{@libPath}/plugins", tasks.completer()
			if @rootPath isnt __dirname and path.existsSync "#{@rootPath}/plugins"
				tasks.push => @loadPlugins "#{@rootPath}/plugins", tasks.completer()
			
			# Execute the loading asynchronously
			tasks.async()

		# Specific Load
		else
			logger.log 'debug', "Plugins loading for #{pluginsPath}"
			util.scandir(
				# Path
				pluginsPath,

				# Skip files
				false,

				# Handle directories
				(fileFullPath,fileRelativePath,_nextFile) =>
					# Prepare
					return nextFile(null,false)  if fileFullPath is pluginsPath
					nextFile = (err,skip) =>
						if err
							logger.log 'warn', "Failed to load the plugin #{loader.pluginName}. The error follows"
							@error err
						_nextFile(null,skip)

					# Prepare
					loader = new PluginLoader dirPath: fileFullPath, docpad: docpad
					pluginName = loader.pluginName
					enabled = (
						(@config.enableUnlistedPlugins  and  @config.enabledPlugins[pluginName]? is false)  or
						@config.enabledPlugins[pluginName] is true
					)

					# Check
					unless enabled
						# Skip
						logger.log 'debug', "Skipping plugin #{pluginName}"
						return nextFile(null,true)
					else
						# Load
						logger.log 'debug', "Loading plugin #{pluginName}"
						loader.exists (err,exists) =>
							return nextFile(err,true)  if err or not exists
							loader.install (err) =>
								return nextFile(err,true)  if err
								loader.require (err) =>
									return nextFile(err,true)  if err
									loader.create {}, (err,pluginInstance) =>
										return nextFile(err,true)  if err
										@pluginsObject[loader.pluginName] = pluginInstance
										@pluginsArray.push pluginInstance
										logger.log 'debug', "Loaded plugin #{pluginName}"
										return nextFile(null,true)
					
				# Next
				(err) =>
					@pluginsArray.sort (a,b) -> a.priority < b.priority
					logger.log 'debug', "Plugins loaded for #{pluginsPath}"
					next(err)
			)
		
		# Chain
		@


	# ---------------------------------
	# Actions

	# Clean the database
	generateClean: (next) ->
		# Before
		@triggerEvent 'cleanBefore', {}, (err) =>
			return next(err)  if err

			# Prepare
			docpad = @
			logger = @logger
			logger.log 'debug', 'Cleaning started'

			# Models
			@cleanModels()
			
			# Async
			tasks = new util.Group (err) ->
				# After
				docpad.triggerEvent 'cleanAfter', {}, (err) ->
					logger.log 'debug', 'Cleaning finished'  unless err
					next(err)
			tasks.total = 6

			# Files
			util.rmdir @outPath, (err,list,tree) ->
				logger.log 'debug', 'Cleaned files'  unless err
				tasks.complete err

			# Layouts
			@layouts.remove {}, (err) ->
				logger.log 'debug', 'Cleaned layouts'  unless err
				tasks.complete err
			
			# Documents
			@documents.remove {}, (err) ->
				logger.log 'debug', 'Cleaned documents'  unless err
				tasks.complete err
			
			# Ensure Layouts
			util.ensurePath @layoutsPath, (err) ->
				logger.log 'debug', 'Ensured layouts'  unless err
				tasks.complete err
			
			# Ensure Documents
			util.ensurePath @documentsPath, (err) ->
				logger.log 'debug', 'Ensured documents'  unless err
				tasks.complete err
		
			# Ensure Public
			util.ensurePath @publicPath, (err) ->
				logger.log 'debug', 'Ensured public'  unless err
				tasks.complete err
		
		# Chain
		@

	# Check if the file path is ignored
	# next(err,ignore)
	filePathIgnored: (fileFullPath,next) ->
		if path.basename(fileFullPath).startsWith('.') or path.basename(fileFullPath).finishesWith('~')
			next null, true
		else
			next null, false
		
		# Chain
		@

	# Parse the files
	generateParse: (next) ->
		# Before
		@triggerEvent 'parseBefore', {}, (err) =>
			return next(err)  if err

			# Requires
			docpad = @
			logger = @logger
			logger.log 'debug', 'Parsing files'

			# Async
			tasks = new util.Group (err) ->
				# Check
				return next(err)  if err
				# Contextualize
				docpad.generateParseContextualize (err) ->
					return next(err)  if err
					# After
					docpad.triggerEvent 'parseAfter', {}, (err) ->
						logger.log 'debug', 'Parsed files'  unless err
						next(err)
			
			# Tasks
			tasks.total = 2

			# Layouts
			util.scandir(
				# Path
				@layoutsPath,

				# File Action
				(fileFullPath,fileRelativePath,nextFile) ->
					# Ignore?
					docpad.filePathIgnored fileFullPath, (err,ignore) ->
						return nextFile(err)  if err or ignore
						layout = docpad.createLayout(
								fullPath: fileFullPath
								relativePath: fileRelativePath
						)
						layout.load (err) ->
							return nextFile(err)  if err
							layout.store()
							nextFile err
					
				# Dir Action
				null,

				# Next
				(err) ->
					logger.log 'warn', 'Failed to parse layouts', err  if err
					tasks.complete err
			)

			# Documents
			util.scandir(
				# Path
				@documentsPath,

				# File Action
				(fileFullPath,fileRelativePath,nextFile) ->
					# Ignore?
					docpad.filePathIgnored fileFullPath, (err,ignore) ->
						return nextFile(err)  if err or ignore
						document = docpad.createDocument(
							fullPath: fileFullPath
							relativePath: fileRelativePath
						)
						document.load (err) ->
							return nextFile err  if err

							# Ignored?
							if document.ignore or document.ignored or document.skip or document.published is false or document.draft is true
								logger.log 'info', 'Skipped manually ignored document:', document.relativePath
								return nextFile()
							
							# Store Document
							document.store()
							nextFile err
				
				# Dir Action
				null,

				# Next
				(err) ->
					logger.log 'warn', 'Failed to parse documents', err  if err
					tasks.complete err
			)

		# Chain
		@
	
	# Generate Parse: Contextualize
	generateParseContextualize: (next) ->
		# Prepare
		docpad = @
		logger = @logger
		logger.log 'debug', 'Parsing files: Contextualizing files'

		# Async
		tasks = new util.Group (err) ->
			return next(err)  if err
			logger.log 'debug', 'Parsing files: Contextualized files'
			next()
		
		# Fetch
		documents = @documents.find({}).sort({'date':-1})
		tasks.total += documents.length

		# Scan all documents
		documents.forEach (document) ->
			document.contextualize tasks.completer()

		# Chain
		@

	# Generate render
	generateRender: (next) ->
		# Prepare
		docpad = @
		logger = @logger
		logger.log 'debug', 'Rendering files'

		# Async
		tasks = new util.Group (err) ->
			return next(err)  if err
			# After
			docpad.triggerEvent 'renderAfter', {}, (err) ->
				logger.log 'debug', 'Rendered files'  unless err
				next(err)
		
		# Prepare site data
		Site = site =
			date: new Date()

		# Prepare template data
		documents = @documents.find({}).sort({'date':-1})
		templateData = 
			require: require
			documents: documents
			document: null
			Documents: documents
			Document: null
			site: site
			Site: site
			blocks:
				scripts: []
				styles: []

		# Before
		@triggerEvent 'renderBefore', {documents,templateData}, (err) ->
			return next(err)  if err
			# Render documents
			tasks.total += documents.length
			documents.forEach (document) ->
				templateData.document = templateData.Document = document
				document.render(
					templateData
					tasks.completer()
				)

		# Chain
		@

	# Write files
	generateWriteFiles: (next) ->
		# Prepare
		logger = @logger
		logger.log 'debug', 'Writing files'

		# Write
		util.cpdir(
			# Src Path
			@publicPath,
			# Out Path
			@outPath
			# Next
			(err) ->
				logger.log 'debug', 'Wrote files'  unless err
				next(err)
		)

		# Chain
		@

	# Write documents
	generateWriteDocuments: (next) ->
		# Prepare
		logger = @logger
		outPath = @outPath
		logger.log 'debug', 'Writing documents'

		# Async
		tasks = new util.Group (err) ->
			logger.log 'debug', 'Wrote documents'  unless err
			next(err)

		# Find documents
		@documents.find {}, (err,documents,length) ->
			# Error
			return tasks.exit err  if err

			# Cycle
			tasks.total += length
			documents.forEach (document) ->
				# Generate path
				fileFullPath = "#{outPath}/#{document.url}" #relativeBase+'.html'
				
				# Ensure path
				util.ensurePath path.dirname(fileFullPath), (err) ->
					# Error
					return tasks.exit err  if err

					# Write document
					logger.log 'debug', "Writing file #{document.relativePath}, #{document.url}"
					fs.writeFile fileFullPath, document.contentRendered, (err) ->
						tasks.complete err

		# Chain
		@

	# Write
	generateWrite: (next) ->
		# Before
		@triggerEvent 'writeBefore', {}, (err) =>
			return next(err)  if err
				
			# Prepare
			docpad = @
			logger = @logger
			logger.log 'debug', 'Writing everything'

			# Async
			tasks = new util.Group (err) ->
				return next(err)  if err
				# After
				docpad.triggerEvent 'writeAfter', {}, (err) ->
					logger.log 'debug', 'Wrote everything'  unless err
					next(err)
			tasks.total = 2

			# Files
			@generateWriteFiles tasks.completer()
			
			# Documents
			@generateWriteDocuments tasks.completer()

		# Chain
		@

	# Generate
	generateAction: (next) ->
		# Before
		@triggerEvent 'generateBefore', {}, (err) =>
			return next(err)  if err
			
			# Prepare
			docpad = @
			logger = @logger
			queryEngine = require('query-engine')  unless queryEngine

			# Clear
			if @generateTimeout
				clearTimeout(@generateTimeout)
				@generateTimeout = null

			# Check
			if @generating
				logger.log 'notice', 'Generate request received, but we are already busy generating... waiting...'
				@generateTimeout = setTimeout(
					=>
						@generateAction next
					1000
				)
			else
				logger.log 'info', 'Generating...'
				@generating = true

			# Continue
			path.exists @srcPath, (exists) ->
				# Check
				if exists is false
					return next new Error 'Cannot generate website as the src dir was not found'
				# Generate Clean
				docpad.generateClean (err) ->
					return next(err)  if err
					# Generate Parse
					docpad.generateParse (err) ->
						return next(err)  if err
						# Generate Render (First Pass)
						docpad.generateRender (err) ->
							return next(err)  if err
							# Generate Render (Second Pass)
							docpad.generateRender (err) ->
								return next(err)  if err
								# Generate Write
								docpad.generateWrite (err) ->
									# Before
									docpad.triggerEvent 'generateAfter', {}, (err) =>
										return next(err)  if err
										# Generated
										growl.notify (new Date()).toLocaleTimeString(), title: 'Website Generated'
										logger.log 'info', 'Generated'
										docpad.generating = false
										next()

		# Chain
		@

	# Watch
	# NOTE: Watching a directory and all it's contents (including subdirs and their contents) appears to be quite expiremental in node.js - if you know of a watching library that is quite stable, then please let me know - b@lupton.cc
	watchAction: (next) ->
		# Prepare
		logger = @logger
		docpad = @
		watch = require 'nodewatch'  unless watch
		logger.log 'Watching setup starting...'
		watch.setMaxListeners(0)

		# Scan the directories inside the srcPath
		util.scandir(
			# Path
			docpad.srcPath,

			# Handle files
			false

			# Handle directories
			(fileFullPath,fileRelativePath,nextFile) =>
				# Watch the src directory
				watch.add(fileFullPath)

				# Continue
				nextFile()

			# Next
			(err) =>
				# Watch the src directory
				watch.add(docpad.srcPath)

				# Changer
				watch.onChange (file,prev,curr) ->
					docpad.generateAction ->
						logger.log 'Regenerated due to file watch at '+(new Date()).toLocaleString()

				# Log
				logger.log 'Watching setup'

				# Next
				next()
		)

		# Chain
		@

	# Skeleton
	skeletonAction: (next) ->
		# Prepare
		logger = @logger
		docpad = @
		skeleton = @skeleton
		destinationPath = @rootPath

		# Copy
		path.exists docpad.srcPath, (exists) =>
			# Check
			if exists
				logger.log 'notice', 'Cannot place skeleton as the desired structure already exists'
				return next()
			
			# Initialise Skeleton
			logger.log 'info', "About to initialise the skeleton [#{skeleton}] to [#{destinationPath}]"
			@initialiseSkeleton skeleton, destinationPath, (err) ->
				return next(err)

		# Chain
		@

	# Server
	serverAction: (next) ->
		# Before
		@triggerEvent 'serverBefore', {@server}, (err) =>
			return next(err)  if err

			# Prepare
			logger = @logger
			express = require 'express'			unless express

			# Server
			if @server
				listen = false
			else
				listen = true
				@server = express.createServer()

			# Configuration
			@server.configure =>
				if listen
					# POST Middleware
					@server.use express.bodyParser()
					@server.use express.methodOverride()

					# Router Middleware
					@server.use @server.router

				# Static
				if @maxAge
					@server.use express.static @outPath, maxAge: @maxAge
				else
					@server.use express.static @outPath
				
				# Routing
				@server.use (req,res,next) =>
					@documents.findOne {urls:{'$in':req.url}}, (err,document) =>
						if err
							@error err
							res.send(err.message, 500)
						else if document
							res.send(document.contentRendered)
						next(err)

				# 404 Middleware
				if listen
					@server.use (req,res,next) ->
						res.send(404)
					
			# Route something
			@server.get /^\/docpad/, (req,res) ->
				res.send 'DocPad!'
			
			# Start server listening
			if listen
				@server.listen @port
				logger.log 'info', 'Express server listening on port', @server.address().port, 'and directory', @outPath

			# After
			@triggerEvent 'serverAfter', {@server}, (err) ->
				logger.log 'debug', 'Server setup'  unless err
				next(err)

		# Chain
		@

# API
docpad =
	Docpad: Docpad
	createInstance: (config) ->
		return new Docpad(config)

# Export
module.exports = docpad
